%{
#include <stdio.h>
#include <string.h>
#include "assembler-const.h"
#include "assembler-yacc.tab.h"
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_stack_ptr;

char currentfilename[256];
char LineBuf[LINEBUFSIZE];
int TokPos;
int LineNo=0;
int InitFlag=0;
%}
WS    [ \t]*
DEC   [0-9]+
HEX \$[0-9A-Fa-f]+

GREGISTER		g[0-1]
RREGISTER		r[0-7]
INCGREG		g[0-1]\+{2}
INCRREG		r[0-8]\+{2}
DECGREG		g[0-1]\-{2}
DECRREG		r[0-8]\-{2}
PC 			pc
BS 			bs
SR 			sr
SP 			sp
FIFO 			fifo
COMMENT 		;[^\n]*
AND 			and
OR  			or
XOR 			xor
ADD			add|\+
SUB 			sub|\-
RTN                     rtn
RTI                     rti
SHIFTL  		<<
SHIFTR  		>>
MEM                     mem
CALL                    call
LCALL                   lcall
JUMP   			jump
JUMPLZ 			jumplz
JUMPNZ			jumpnz
MASK                    mask
SHIFTFIFO 		shiftfifo
READFIFO                readfifo
POPFIFO			popfifo
PUSH			push
POP  			pop
NOP    	  		nop
RESETFIFO 		resetfifo
MUL     		\*
DIV     		\/
POW     		\^
LABEL   		[^ \t\n\[\]:\.;\+\-=\(\)\^\*\/\$,]+
WORD			\.word
SETADR  		\#set[ ]+address
INCLUDE 		\#include
FILE    		[^ \n\t]+
DEFINE  		\#define
HUFFTABLE               \#hufftable
ENDNODE                 endnode
JUMPNODE                jumpnode

%x endline code include define adr scan newline 

%%
%{
	if (InitFlag==0)
	{	
		BEGIN scan;
		InitFlag=1;
        }
%}

<scan>^.*$ { 
       BEGIN newline;
       TokPos=0;
       strcpy(LineBuf,yytext);
       yyless(0);
}

<newline>{LABEL}:	{  
			BEGIN code;
			TokPos+=yyleng; 
			strcpy(yylval.str,yytext);
			yylval.str[strlen(yylval.str)-1]='\0';
			return DEFLABEL;
			}

<newline>{INCLUDE}	{
			TokPos+=yyleng; 
			BEGIN include;
			return INCLUDE;
			}

<newline>{DEFINE}	{
			TokPos+=yyleng; 
			BEGIN define;
			return DEFINE;
			}

<newline>{SETADR}	{
			TokPos+=yyleng; 
			BEGIN code;
			return SETADR;
			}

<newline>{HUFFTABLE}    {
                        TokPos+=yyleng;
                        BEGIN endline;
                        return HUFFTABLE;
                        }


<newline>{COMMENT}     { strcpy(yylval.str,yytext);
                         BEGIN endline;
                         return NLCOMMENT;
                       }

<newline>[ ]+	{
		TokPos+=yyleng;
		BEGIN code;
		}

<newline>[\t]	{
		TokPos+=TABWIDTH;
		BEGIN code;
		}

<*>{COMMENT} { strcpy(yylval.str,yytext);
	       BEGIN endline;
               return COMMENT;
}

<include>{FILE}	{
		TokPos+=yyleng;
		InitFlag=0;
		strcpy(yylval.str,yytext);
                return INCLUDEFILE;
		}

<define>{LABEL}	{
		TokPos+=yyleng;
		BEGIN code;
		strcpy(yylval.str,yytext);
                return LABEL;
		}

<code>{WORD}	{
		TokPos+=yyleng;
		BEGIN adr;
                return WORD;
		}

<adr,code>{DEC}	{ 
		TokPos+=yyleng;
        	yylval.num=atoi(yytext);
		return NUM;
		}

<adr,code>{HEX}	{ 
		TokPos+=yyleng;
        	yylval.num=hextodec(yytext+1);
		return NUM;
		}

<code>{INCRREG}	{ 
		TokPos+=yyleng;
                yylval.num=atoi(yytext+1)+NoOfGRegs;
	        return INCREG;
		}

<code>{INCGREG}	{ 
		TokPos+=yyleng;
                yylval.num=atoi(yytext+1);
	        return INCREG;
		}

<code>{DECRREG}	{ 
		TokPos+=yyleng;
                yylval.num=atoi(yytext+1)+NoOfGRegs;
	        return DECREG;
		}

<code>{DECGREG}	{ 
		TokPos+=yyleng;
                yylval.num=atoi(yytext+1);
	        return DECREG;
		}

<code>{GREGISTER} { 
		TokPos+=yyleng;
                                        yylval.num=atoi(yytext+1);
		return REG;
		}

<code>{RREGISTER} { 
		TokPos+=yyleng;
        	                    yylval.num=atoi(yytext+1)+NoOfGRegs;
		return REG;
		}

<code>{SP}	{ 
		TokPos+=yyleng;
		return SP;
		}

<code>{SR}	{ 
		TokPos+=yyleng;
		return SR;
		}

<code>{PC}	{ 
		TokPos+=yyleng;
		return PC;
		}

<code>{BS}	{ 
		TokPos+=yyleng;
		return BS;
		}

<code>{FIFO}	{ 
		TokPos+=yyleng;
		return FIFO;
		}

<code>{AND}	{ 
		TokPos+=yyleng;
	        yylval.num=AND_calc;
	        return OPERATOR;
		}

<code>{OR}	{ 
		TokPos+=yyleng;
	        yylval.num=OR_calc;
	        return OPERATOR;
		}

<code>{XOR}	{ 
		TokPos+=yyleng;
	        yylval.num=XOR_calc;
	        return OPERATOR;
		}

<code>{SHIFTL}	{ 
		TokPos+=yyleng;
	        yylval.num=SLL_calc;
	        return OPERATOR;
		}

<code>{SHIFTR}	{ 
		TokPos+=yyleng;
	        yylval.num=SLR_calc;
	        return OPERATOR;
		}

<code>{MEM}	{ 
		TokPos+=yyleng;
	        return MEMOP;
		}

<code>{PUSH}	{ 
		TokPos+=yyleng;
	        return PUSH;
		}

<code>{MASK}	{ 
		TokPos+=yyleng;
	        return MASK;
		}

<code>{POP}	{ 
		TokPos+=yyleng;
	        return POP;
		}

<code>{NOP}	{ 
		TokPos+=yyleng;
	        return NOP;
		}

<code>{RESETFIFO} { 
		TokPos+=yyleng;
	        return RESETFIFO;
		}


<code>{JUMP}	{ 
		TokPos+=yyleng;
	        return JUMP;
		}

<code>{CALL}	{ 
		TokPos+=yyleng;
	        return CALL;
		}

<code>{LCALL}	{ 
		TokPos+=yyleng;
	        return LCALL;
		}

<code>{JUMPLZ}	{ 
		TokPos+=yyleng;
	        return JUMPLZ;
		}

<code>{JUMPNZ}	{ 
		TokPos+=yyleng;
	        return JUMPNZ;
		}

<code>{POPFIFO}	{ 
		TokPos+=yyleng;
	        return POPFIFO;
		}

<code>{READFIFO} { 
		TokPos+=yyleng;
	        return READFIFO;
		}

<code>{SHIFTFIFO} { 
		TokPos+=yyleng;
	        return SHIFTFIFO;
		}

<code>{RTN}	{ 
		TokPos+=yyleng;
	        return RTN;
		}

<code>{ENDNODE}	{ 
		TokPos+=yyleng;
	        return NODEEND;
		}

<code>{JUMPNODE} { 
		 TokPos+=yyleng;
	         return NODEJUMP;
		 }

<code>{RTI}	{ 
		TokPos+=yyleng;
	        return RTI;
		}

<adr,code>{ADD}	{ 
		TokPos+=yyleng;
	        yylval.num=ADD_calc;
	        return OPERATOR;
		}

<adr,code>{SUB}	{ 
		TokPos+=yyleng;
	        yylval.num=SUB_calc;
	        return OPERATOR;
		}

<adr,code>{MUL}	{ 
		TokPos+=yyleng;
	        yylval.num=MUL_calc;
	        return OPERATOR;
		}

<adr,code>{DIV}	{ 
		TokPos+=yyleng;
	        yylval.num=DIV_calc;
	        return OPERATOR;
		}


<adr,code>{POW}	{ 
		TokPos+=yyleng;
	        yylval.num=POW_calc;
	        return OPERATOR;
		}

<adr,code>{LABEL} {
		TokPos+=yyleng;
		strcpy(yylval.str,yytext);
                                        return LABELREF;
		}

<include,define,endline,adr,code>[ ]+ { TokPos+=yyleng; }
<include,define,endline,adr,code>[\t]+ { TokPos+=(yyleng*TABWIDTH); }
		
<endline,adr,code>\n   { BEGIN scan; 
                         LineNo++; 
                        }
		
<*><<EOF>> {
  if ( --include_stack_ptr < 0 )
    {
      yyterminate();
    }
  
  else
    {
      yy_delete_buffer( YY_CURRENT_BUFFER );
      yy_switch_to_buffer(include_stack[include_stack_ptr] );
    }
  strcpy(yylval.str,"; end of include");
  return COMMENT;
}

<*>. { TokPos+=1; return yytext[0]; }

<*>\n { BEGIN scan;
     LineNo++; 
}

%%

void startnewfile(char *filename)
{
  if ( include_stack_ptr >= MAX_INCLUDE_DEPTH )
    yyerror("Includes nested too deeply");
  
  include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;

  sprintf(currentfilename,"%s.asm",filename);
  yyin = fopen(currentfilename, "r" );
  
  if ( ! yyin )
    yyerror("Could not open include file");
  
  yy_switch_to_buffer( yy_create_buffer( yyin, YY_BUF_SIZE ) );
  
  BEGIN 0;
}
