%{
#include <stdio.h>
#include "assembler-const.h"
int Address=0;
int HuffmanBaseAdr;
static int no=0;
char tag[LABELSIZE];
char symline[LINEBUFSIZE];
char reg1[10];
char reg2[10];
char reg3[10];
extern FILE *symout;
extern int Pass;

struct labeltabletype
{
  char name[LABELSIZE];
  int value;
} labeltable[MAXLABLES];

void HandleComments(char *comment);
void HandleNLComments(char *comment);
void HandleConstant(int adr, int value, char *label);
void HandleSetAddress(int newadr,int oldadr);
void HandleLabel(char *label, int value);
void HandleDefines(char *label, int value);
void HandleInclude(char *filename);
char* regstr(int regnr, char *reg);
%}


%union {
  int num;
  char str[LABELSIZE];
}

%start file
%left '-' '+'
%left '*' '/'
%left '^'
%token <num> REG OPERATOR NUM WORD SETADR VAR MEMOP PC SP FIFO JUMP JUMPLZ JUMPNZ SHIFTFIFO POPFIFO PUSH POP RESETFIFO NOP DEFINE RTN READFIFO CALL INCREG NODEJUMP NODEEND HUFFTABLE DECREG MASK SR RTI BS LCALL 
%token <str> COMMENT LABELREF INCLUDE INCLUDEFILE DEFLABEL LABEL NLCOMMENT
%type <num> imm constant binary address set register jump fifo stakop unary call code incdec masking returnintr lcall huffmantable

%%

file :/* empty */
     | file code { tag[0]='\0'; symline[0]='\0'; }
     ; 

code : line {
                   fprintf(symout,"%s\n",symline); 
                 }
     | line COMMENT {
       HandleComments($2);
       fprintf(symout,"%s\n",symline); 

                         }
     ;

line : binary    { Address++; }
     | set       { Address++; }
     | mem       { Address++; }
     | jump      { Address++; }
     | masking   { Address++; }
     | call      { Address++; }
     | lcall     { Address += 2; }
     | fifo      { Address++; }
     | stakop    { Address++; }
     | returnintr { Address++; }
     | unary     { Address++; }
     | incdec    { Address++; }
     | constant  { HandleConstant(Address,$1,tag); 
                   Address++;
		   makeopcode_constant($1);
                 }
     | address   { HandleSetAddress($1,Address);
		   makeopcode_zeros(Address,$1);
                   Address=$1; 
                 }
     | NLCOMMENT { HandleNLComments($1); }
     | COMMENT   { HandleComments($1); }
     | DEFLABEL  { HandleLabel($1,Address);    }
     | DEFINE LABEL '=' imm { HandleDefines($2,$4); }
     | INCLUDE INCLUDEFILE { 
                HandleInclude($2); 
		startnewfile($2);
                }
     | HUFFTABLE { HuffmanBaseAdr=Address; sprintf(symline,"#HUFFTABLE"); }
     | huffmantable { Address++; }
     ;

huffmantable :
             NODEJUMP '(' imm ',' imm ')' 
             {
	       sprintf(symline,"     [%4X]          JUMPNODE (%d,%d) [ %s ]",
		       Address, $3, $5-HuffmanBaseAdr, tag);
	       makeopcode_jumpnode($3-1, $5-HuffmanBaseAdr);
	     }
             |
             NODEEND '(' imm ',' imm ',' imm ')' 
             {
                 sprintf(symline,"     [%4X]          ENDNODE (%d,%d,%d)",  
			 Address, $3, $5, $7);
		 makeopcode_endnode($3-1,$5,$7);
             }
             ;

stakop : PUSH '(' register ')' {
                 sprintf(symline,"     [%4X]          PUSH ( %s )",  
			 Address, regstr($3,reg1));
		  makeopcode_push($3);  }
       | register '=' POP '(' ')' {
                 sprintf(symline,"     [%4X]          %s = POP()",  
			 Address, regstr($1,reg1));
		  makeopcode_pop($1);  }
       | RTN  {
                 sprintf(symline,"     [%4X]          RTN",  
			 Address);
        	  makeopcode_pop(REGPC);  }
       ;

returnintr : RTI  {
                 sprintf(symline,"     [%4X]          RTI",  
			 Address);
        	  makeopcode_rti();
                  }
      ;
   
unary : NOP {
             sprintf(symline,"     [%4X]          NOP", Address);
		  makeopcode_nop();  }
      | RESETFIFO {
             sprintf(symline,"     [%4X]          RESETFIFO", Address);
	     makeopcode_resetfifo();  }
      ;


fifo : register '=' POPFIFO '(' register ')' {
                  sprintf(symline,"     [%4X]          %s = POPFIFO ( %s )",  
			  Address, regstr($1,reg1), regstr($5,reg2));
		  makeopcode_popfifo($1,$5);  }
     |  register '=' POPFIFO '(' imm ')' {
                  sprintf(symline,"     [%4X]          %s = POPFIFO ( %d )",  
			   Address, regstr($1,reg1), $5);
		  makeopcode_popfifoi($1,$5);  }
     | SHIFTFIFO '(' register ')' {
                  sprintf(symline,"     [%4X]          SHIFTFIFO ( %s )",  
			  Address, regstr($3,reg1));
		  makeopcode_popfifo(REGIMM,$3);  }
     | SHIFTFIFO '(' imm ')' {
                  sprintf(symline,"     [%4X]          SHIFTFIFO ( %d )",  
			  Address, $3);
		  makeopcode_popfifoi(REGIMM,$3);  }
     ;

jump : JUMP imm {
                  sprintf(symline,"     [%4X]          JUMP PC + %d [ %s ]",  
			  Address, $2-Address-1,tag);
		  makeopcode_jump($2-Address-1);  }
     | JUMPLZ imm {
                  sprintf(symline,"     [%4X]          JUMPLZ PC + %d [ %s ]",  
			  Address, $2-Address-1,tag);
		  makeopcode_jumplz($2-Address-1);  }
     | JUMPNZ imm {
                  sprintf(symline,"     [%4X]          JUMPNZ PC + %d [ %s ]",  
			  Address, $2-Address-1,tag);
		  makeopcode_jumpnz($2-Address-1);  }


call : CALL imm {
                  sprintf(symline,"     [%4X]          CALL PC + %d [ %s ]",  
			  Address, $2-Address-1,tag);
		  makeopcode_call($2-Address-1);  }
     ;

lcall : LCALL imm {
                  sprintf(symline,"     [%4X]          LCALL PC = %d [ %s ]",  
			  Address, $2,tag);
		  makeopcode_lcall($2);  }
     ;

masking : MASK '(' INCREG ',' imm ')' {
                 sprintf(symline,"     [%4X]          MASK ( %s++, %d )",  
			 Address, regstr($3,reg1),$5);
		  makeopcode_mask($3,$5);  }
        ;

mem      : register '=' MEMOP '(' register OPERATOR register ')' {
                  sprintf(symline,"     [%4X]          %s = MEM(%s + %s)",
			  Address, regstr($1,reg1), 
			  regstr($5,reg2), regstr($7,reg3));
		  makeopcode_load($1,$5,$7); } 
         | register '=' MEMOP '(' register OPERATOR imm ')' {
	   if (strcmp(tag,"")==0)
	     sprintf(symline,"     [%4X]          %s = MEM(%s %s %d)",
		     Address, regstr($1,reg1), regstr($5,reg2),
		     getoperatorstr($6), $7);
	   else
	     sprintf(symline,"     [%4X]          %s = MEM(%s %s %d [ %s ])",
		     Address, regstr($1,reg1), regstr($5,reg2),
		     getoperatorstr($6), $7,tag);
	   makeopcode_loadi($1,$5,$7,$6); } 
         | register '=' MEMOP '(' register ')' {
	   sprintf(symline,"     [%4X]          %s = MEM(%s)",
		   Address, regstr($1,reg1), regstr($5,reg2));
	   makeopcode_load($1,$5,REGIMM); } 
         | register '=' MEMOP '(' imm ')' {
	   sprintf(symline,"     [%4X]          %s = MEM(PC+%d [ %s ])",
		   Address, regstr($1,reg1),$5-Address-1, tag);
	   makeopcode_loadi($1,REGPC,$5-Address-1,ADD_calc); } 
         |  MEMOP '(' register OPERATOR register ')' '=' register {
	   sprintf(symline,"     [%4X]          MEM(%s %s %s) = %s",
		   Address, regstr($3,reg1),getoperatorstr($4), 
		   regstr($5,reg2), regstr($8,reg3));
	   makeopcode_store($3,$5,$8); }
         |  MEMOP '(' register OPERATOR imm ')' '=' register {
	   if (strcmp(tag,"")==0)
 	     sprintf(symline,"     [%4X]          MEM(%s %s %d) = %s",
		     Address, regstr($3,reg1), getoperatorstr($4), 
		     $5, regstr($8,reg2));
	   else
 	     sprintf(symline,"     [%4X]          MEM(%s %s %d [ %s ]) = %s",
		     Address, regstr($3,reg1), getoperatorstr($4), 
		     $5, tag,regstr($8,reg2));
	   makeopcode_storei($3,$5,$8,$4); }
         |  MEMOP '(' register ')' '=' register {
                  sprintf(symline,"     [%4X]          MEM(%s) = %s",
			  Address, regstr($3,reg1),regstr($6,reg2));
		  makeopcode_store($3,REGIMM, $6,ADD_calc); }
         |  MEMOP '(' imm ')' '=' register {
                  sprintf(symline,"     [%4X]          MEM(PC+%d [ %s ]) = %s",
			  Address, $3-Address-1, tag, regstr($6,reg1));
		  makeopcode_storei(REGPC,$3-Address-1,$6,ADD_calc); }
         ;
constant : WORD imm { $$=$2; } ; 

set      : register '=' imm {
                 sprintf(symline,"     [%4X]          %s = %d",Address,regstr($1,reg1),$3);
                 makeopcode_set($1,$3) }
         ;

binary : register '=' register OPERATOR register { 
                     sprintf(symline,"     [%4X]          %s = %s %s %s",
			     Address, regstr($1,reg1), regstr($3,reg2), 
			     getoperatorstr($4),regstr($5,reg3));
		     makeopcode_rrr($1,$3,$4,$5); }
       | register '=' register { 
                     sprintf(symline,"     [%4X]          %s = %s",
			     Address, regstr($1,reg1), regstr($3,reg2));
		     makeopcode_rrr($1,$3,AND_calc,$3); }
       | register '=' register OPERATOR imm { 
                     sprintf(symline,"     [%4X]          %s = %s %s %d",
			     Address, regstr($1,reg1), regstr($3,reg2), 
			     getoperatorstr($4),$5);
		     makeopcode_rri($1,$3,$4,$5); }
       | register '=' READFIFO '(' imm ')' { 
                     sprintf(symline,"     [%4X]          %s = READFIFO ( %d )",
			     Address, regstr($1,reg1),$5);
		     makeopcode_readfifoi($1,$5); }
       | register '=' READFIFO '(' register ')' { 
                     sprintf(symline,"     [%4X]          %s = READFIFO ( %s )",
			     Address, regstr($1,reg1), regstr($5,reg2));
		     makeopcode_readfifo($1,$5); }
       ;

incdec : INCREG {
           sprintf(symline,"     [%4X]          %s++",
		   Address, regstr($1,reg1));
	   makeopcode_inc($1); }
       | DECREG {
           sprintf(symline,"     [%4X]          %s--",
		   Address, regstr($1,reg1));
	   makeopcode_dec($1); }
         
       ;

address  : SETADR '=' imm { $$=$3; };

imm : imm OPERATOR imm { $$=calculate($1,$3,$2); }
    | '(' imm ')' { $$ = $2; }
    | NUM         { $$ = $1; }
    | LABELREF    { sprintf(tag,"%s",$1); $$ = readlabel($1); }           
    ;

register : REG { $$=$1; }
         | PC  { $$=REGPC; }
         | SP  { $$=REGSP; }
         | SR  { $$=REGSR; }
         | FIFO { $$=REGFIFO; }
         | BS { $$=REGBS; }
         ;

%%
	 
void HandleComments(char *comment)
{
  int count=0;
  char empty[COMMENTPLACEMENT+1];
  for (count=0;count<COMMENTPLACEMENT;count++)
    empty[count]=' ';
  empty[count+1]='\0';
  count=0;

  while (symline[count++]!='\0');
  count--;
  if (count<COMMENTPLACEMENT)
    {
      for (;count<COMMENTPLACEMENT;count++)
	symline[count]=' ';
      sprintf(&symline[count]," %s",comment);
    }
  else
    {
      
      
      sprintf(&symline[count],"\n%s %s",empty,comment);
    }
}

void HandleNLComments(char *comment)
{
  sprintf(symline,"%s",comment);
}

void HandleConstant(int Address, int value, char *label)
{
  if (strcmp(label,"")==0)
    sprintf(symline,"     [%4X] .WORD    %d ",Address,value);
  else
    sprintf(symline,"     [%4X] .WORD    %d  [%s]",Address,value,label);
  sprintf(label,""); 
}

void HandleSetAddress(int newadr, int oldadr)
{
  if (newadr<oldadr)
    {
      fprintf(stderr,"Current address is %d\n",oldadr);
      yyerror("Sorry, the assembler does not support a '#set address' in negative direction");
    }
  sprintf(symline,"#SET ADDRESS = %d",newadr);
}

void HandleLabel(char *label, int value)
{
  int x;
  if (Pass==FIRST)
    for (x=0;x<no;x++)
      if (strcmp(labeltable[x].name,label)==0)
	yyerror("Duplicant labels");
  if (Pass==FIRST)
    {
      strcpy(labeltable[no].name,label);
      labeltable[no].value=value;
      no++;
      if (no==MAXLABLES)
	yyerror("Sorry too many lables !!");
    }
  
  sprintf(symline,"%s:",label);
}

void HandleDefines(char *label, int value)
{
  int x;
  if (Pass==FIRST)
  for (x=0;x<no;x++)
    if (strcmp(labeltable[x].name,label)==0)
      yyerror("Duplicant labels");
  if (Pass==FIRST)
    {
      strcpy(labeltable[no].name,label);
      labeltable[no].value=value;
      no++;
      if (no==MAXLABLES)
	yyerror("Sorry too many lables !!");
    }

  sprintf(symline,"#DEFINE %s = %d",label,value);      
}


int readlabel(char *label)
{
  int no=0;

  if (Pass==FIRST)
      return 0;
  else if (Pass==SECOND)
    {
      while ((labeltable[no].name!='\0') && (no<MAXLABLES))
	{
	  if (strcmp(labeltable[no].name,label)==0)
	    return labeltable[no].value;
	  no++;
	}
      yywarning("Could not find label");
      return 0;
    }
  else
    {
      fprintf(stderr,"Can not determine pass level \n");
      exit(1);
    }
}

void HandleInclude(char *file)
{
  sprintf(symline,"#INCLUDE %s",file);
}

char *regstr(int regnr, char *reg)
{
  if ((regnr>=0) && (regnr<NoOfGRegs))
    sprintf(reg,"G%d",regnr);
  else if ((regnr>=(NoOfGRegs)) && 
	     (regnr<(NoOfGRegs+NoOfRRegs)))
    sprintf(reg,"R%d",regnr-NoOfGRegs);
  else
    switch (regnr)
      {
      case REGBS: sprintf(reg,"BS");
	break;
      case REGSR: sprintf(reg,"SR");
	break;
      case REGPC: sprintf(reg,"PC");
	break;
      case REGSP: sprintf(reg,"SP");
	break;
      case REGFIFO: sprintf(reg,"FIFO");
	break;
      default:
	yyerror("unknown register");
      }
  return reg;
}


